"""
    线程锁装饰器，直接锁整个函数
"""
import threading
from typing import Callable


class FuncLockDecorator:
    """
        通过线程执行函数，简化创建线程
    """

    def __init__(self, timeout: int = -1):
        self.timeout = timeout
        self.lock = threading.RLock()  # 包住整个函数的，所以在实例化里面

    def __call__(self, func: Callable):
        def _wrapper(*args, **kwargs):

            self.lock.acquire(timeout=self.timeout)

            try:
                return func(*args, **kwargs)
            finally:
                self.lock.release()

        return _wrapper



"""
    spider 结束中间件，最后一个运行

    仅用作提示
"""
from loguru import logger
from palp.middleware.middleware_spider import SpiderMiddleware


class SpiderEndMiddleware(SpiderMiddleware):
    """
        spider 资源回收中间件
    """

    def spider_close(self, spider):
        """
        关闭所有创建的连接

        :param spider:
        :return:
        """
        logger.debug("爬虫已停止")



"""
    抛出错误
"""


class NotGeneratorFunctionError(Exception):
    """
        函数是含有 yield 的函数
    """

    def __init__(self, *args):
        super(NotGeneratorFunctionError, self).__init__(*args)


class SpiderHasExistsError(Exception):
    """
        spider 已存在错误
    """

    def __init__(self, *args):
        super(SpiderHasExistsError, self).__init__(*args)


class DropRequestException(Exception):
    """
        丢弃请求
    """

    def __init__(self, *args):
        super(DropRequestException, self).__init__(*args)


class DropItemException(Exception):
    """
        丢弃 item
    """

    def __init__(self, *args):
        super(DropItemException, self).__init__(*args)


class NotStrictItemFieldException(Exception):
    """
        非 StrictItem 字段错误
    """

    def __init__(self, *args):
        super(NotStrictItemFieldException, self).__init__(*args)



"""
    修改 loguru 的 logger
"""
import loguru


def debug(*args, sep=' '):
    if len(args) == 1:
        loguru.logger.debug(args[0])
    else:
        loguru.logger.debug(f'{sep}'.join(args))


def info(*args, sep=' '):
    if len(args) == 1:
        loguru.logger.info(args[0])
    else:
        loguru.logger.info(f'{sep}'.join(args))


def warning(*args, sep=' '):
    if len(args) == 1:
        loguru.logger.warning(args[0])
    else:
        loguru.logger.warning(f'{sep}'.join(args))


def error(*args, sep=' '):
    if len(args) == 1:
        loguru.logger.error(args[0])
    else:
        loguru.logger.error(f'{sep}'.join(args))


def exception(arg):
    loguru.logger.exception(arg)



"""
    等待 spider 各项功能执行结束
"""


class SpiderWaitDecorator:
    """
        等待 spider controller、item controller 执行完毕
    """

    def __call__(self, func):
        def _wrapper(*args, **kwargs):
            # 先执行函数
            try:
                func(*args, **kwargs)
            finally:
                # 获取 spider
                spider = args[0]

                # 等待 spider 控制器执行结束
                spider.wait_spider_controller_done()

                # 等待 spider item 分发执行完毕
                spider.wait_item_controller_done()

        return _wrapper



"""
    spider 中间件装饰器
"""
from palp import settings
from loguru import logger


class SpiderMiddlewareDecorator:
    """
        spider 中间件装饰器
    """

    def __call__(self, func):
        def _wrapper(*args, **kwargs):
            spider = args[0]  # 其实就是获取 self
            middlewares = spider.__class__.SPIDER_MIDDLEWARE  # 获取 middleware

            for middleware in middlewares:
                middleware.spider_start(spider)

            try:
                func(*args, **kwargs)
            except Exception as e:
                if settings.SPIDER_STOP_ON_ERROR:
                    spider.stop_all_spider_controller()
                    logger.exception(e)
                else:
                    for middleware in middlewares:
                        middleware.spider_error(spider, e)
            finally:
                for middleware in middlewares:
                    try:
                        middleware.spider_close(spider)
                    except Exception as e:
                        logger.exception(e)

        return _wrapper



"""
    request 失败回收中间件
"""
from palp import settings
from palp.middleware.middleware_request import RequestMiddleware


class RequestRecycleMiddleware(RequestMiddleware):
    """
        失败请求回收中间件
    """

    def request_failed(self, spider, request):
        """
        回收失败的请求

        注意：开启 REQUEST_FAILED_SAVE 才会启用

        :param spider:
        :param request:
        :return:
        """
        from palp.conn import redis_conn

        if redis_conn is None and settings.SPIDER_TYPE == 1:
            return

        redis_conn.sadd(settings.REDIS_KEY_QUEUE_BAD_REQUEST, request.to_json())



"""
    请求的处理
"""
import json
import time
import dill
import base64
import random
import urllib3
from palp import settings
from typing import Callable
from urllib.parse import urlparse
from palp.network.response import Response
from palp.tool.user_agent import random_ua
from requests.cookies import RequestsCookieJar
from palp.tool.short_module import import_module
from urllib3.exceptions import InsecureRequestWarning

# 禁用不安全警告
urllib3.disable_warnings(InsecureRequestWarning)

# 需要 pickle 的属性
pickle_attr = [
    'meta', 'downloader', 'downloader_parser', 'cookie_jar', 'command', 'jump_spider',
    'jump_spider_kwargs', 'jump_request_middleware'
]


class Request:
    # requests 模块所需的
    __REQUEST_ATTRS__ = [
        'url',
        'method',
        'params',
        'data',
        'headers',
        'cookies',
        'files',
        'auth',
        'timeout',
        'allow_redirects',
        'proxies',
        'hooks',
        'stream',
        'verify',
        'cert',
        'json',
    ]

    # 框架运行所需的
    __PALP_ATTRS__ = [
        'downloader',
        'downloader_parser',
        'filter_repeat',
        'keep_session',
        'keep_cookie',
        'callback',
        'cookie_jar',
        'priority',
        'command',
        'meta',
        'jump_spider',
        'jump_spider_kwargs',
        'jump_request_middleware',
    ]

    # 下载器
    DOWNLOADER = None
    DOWNLOADER_PARSER = None

    def __new__(cls, *args, **kwargs):
        """
        导入下载器

        注意：Request 写了很多方法，但是 LoadRequest 是基于 Request 这个父类判断的，所以这里不使用 cls.xxx

        :param args:
        :param kwargs:
        """
        if Request.DOWNLOADER is None:
            Request.from_settings()

        return object.__new__(cls)

    @classmethod
    def from_settings(cls):
        """
        导入 downloader、downloader_parser

        :return:
        """
        Request.DOWNLOADER = import_module(settings.RESPONSE_DOWNLOADER, instantiate=False)[0]
        Request.DOWNLOADER_PARSER = import_module(settings.RESPONSE_DOWNLOADER_PARSER, instantiate=False)[0]

    def __init__(
            self,
            url: str,
            method: str = None,
            params: dict = None,
            data: dict = None,
            headers: dict = None,
            cookies: dict = None,
            timeout: float = None,
            proxies: dict = None,
            json: dict = None,
            meta: dict = None,
            downloader=None,
            downloader_parser=None,
            filter_repeat: bool = False,
            keep_session: bool = False,
            keep_cookie: bool = False,
            callback: Callable = None,
            cookie_jar: RequestsCookieJar = None,
            priority: int = settings.DEFAULT_QUEUE_PRIORITY,
            command: dict = None,
            jump_spider=None,
            jump_spider_kwargs: dict = None,
            jump_request_middleware=None,
            **kwargs
    ):
        """
        requests 参数
        :param url: 请求链接
        :param params: params
        :param data: data
        :param headers: 请求头
        :param cookies: cookies
        :param timeout: 超时时间
        :param proxies: 代理：{'http':'http://ip:port','https':'https://ip:port'}
        :param json: json：也可以使用 data = json.dumps(data)

        Palp 参数
        :param filter_repeat: 是否过滤请求，settings.FILTER_REQUEST 启用生效，默认 False
        :param keep_session: 是否保持 session，默认 False
        :param keep_cookie: 不使用 session 时保持 cookie，默认 False
        :param callback: 回调函数
        :param priority: 启用优先级队列时的优先级，分数越大，优先级约低，默认 settings.DEFAULT_QUEUE_PRIORITY
        :param command: 自定义操作命令，用于自定义 downloader 时使用
        :param downloader: 自定义的下载器（局部）
        :param downloader_parser: 自定义的下载器的解析器（局部）
        :param meta: 自动向下传递参数，临时的直接 xxx=yyy，自动的 meta={'xxx':'yyy'}
        :param jump_spider: 跳转的 spider
        :param jump_spider_kwargs: 跳转的 spider 需要 self.xxx 的参数
        :param jump_request_middleware: 跳转的 spider 需要的请求中间件

        Palp 参数（非用户设置）
        :param cookie_jar: cookie_jar，存储 cookie，这里使用的是 requests 模块的，其它请求的话可以自己提取

        传递的参数
        :param kwargs: 其余需要传递的参数，若参数名不存在则返回 None
        """

        # Request 所需字段
        self._requests_params = {}  # request 参数
        self.meta = meta
        self.command = command
        self.callback = callback
        self.priority = priority
        self.downloader = downloader
        self.cookie_jar = cookie_jar
        self.jump_spider = jump_spider
        self.keep_cookie = keep_cookie
        self.keep_session = keep_session
        self.filter_repeat = filter_repeat
        self.downloader_parser = downloader_parser
        self.jump_spider_kwargs = jump_spider_kwargs
        self.jump_request_middleware = jump_request_middleware

        # requests 请求参数
        self.url = url
        self.data = data
        self.json = json
        self.method = method
        self.params = params
        self.headers = headers
        self.cookies = cookies
        self.timeout = timeout
        self.proxies = proxies

        # 传递参数更新一下
        for key, value in kwargs.items():
            setattr(self, key, value)

    def send(self) -> Response:
        """
        获取响应

        :return: ResponseParser 的解析器
        """
        # 加载默认值
        self.set_default()

        # 判断是否等待
        if isinstance(settings.REQUEST_DELAY, int) or isinstance(settings.REQUEST_DELAY, float):
            time.sleep(settings.REQUEST_DELAY)
        elif isinstance(settings.REQUEST_DELAY, list):
            time.sleep(random.choice([i for i in range(settings.REQUEST_DELAY[0], settings.REQUEST_DELAY[1] + 1)]))

        # cookie_jar 更新传入的 cookie
        self.cookie_jar.update(self.cookies)

        # 获取响应
        response = self.downloader(
            **self._requests_params,
            keep_session=self.keep_session,
            keep_cookie=self.keep_cookie,
            cookie_jar=self.cookie_jar,
            command=self.command
        ).response()

        response_parser = self.downloader_parser(response)  # 解析器解析响应
        if response_parser.cookies:
            self.cookie_jar.update(response_parser.cookies)  # 将响应的 cookie 更新到 cookieJar

        return response_parser

    def set_default(self):
        """
        设置一些默认值

        :return:
        """
        # 判断类型
        if self.method:
            self.method = self.method.upper()
        elif self.data or self.json:
            self.method = 'POST'
        else:
            self.method = 'GET'

        # 设置默认
        if self.meta is None:
            self.meta = {}
        if self.downloader is None:
            self.downloader = self.__class__.DOWNLOADER
        if self.downloader_parser is None:
            self.downloader_parser = self.__class__.DOWNLOADER_PARSER
        if self.cookie_jar is None:
            self.cookie_jar = RequestsCookieJar()
        if self.command is None:
            self.command = {}
        if self.headers is None:
            self.headers = {}
        if self.cookies is None:
            self.cookies = {}
        elif isinstance(self.cookies, tuple):
            self.cookies = dict(self.cookies)
        if self.timeout is None:
            self.timeout = settings.REQUEST_TIMEOUT or 60
        self._requests_params.setdefault('verify', False)

        # 添加 ua（模块在使用 ua 的时候可能访问降速）
        if settings.RANDOM_USERAGENT:
            ua = random_ua()
        else:
            ua = settings.DEFAULT_USER_AGENT

        if not self.headers:
            self.headers.update({'User-Agent': ua})
        elif not self.headers.get('User-Agent'):
            self.headers.update({'User-Agent': ua})
        elif not self.headers.get('user-agent'):
            self.headers.update({'user-agent': ua})

    def to_dict(self) -> dict:
        """
        获取字典形式

        主要作用：快速进行二次请求

        示例：
            request_dict = request.to_dict()
            request_dict[xxx] = xxx # 修改

            yield palp.Request(**request_dict)
        :return:
        """
        request_dict = {}

        # 先取出 callback
        if self.callback and not isinstance(self.callback, str):
            request_dict['callback'] = self.callback.__name__

        # 提取其它参数
        for key, value in self.__dict__.items():
            # _打头的名字 和 无值的忽略
            if key.startswith('_') or not value:
                continue

            # downloader 为默认的，直接忽略
            elif key == 'downloader' and value == self.__class__.DOWNLOADER:
                continue

            # downloader_parser 为默认的，直接忽略
            elif key == 'downloader_parser' and value == self.__class__.DOWNLOADER_PARSER:
                continue

            # 序列化对象
            elif key in pickle_attr:
                request_dict[key] = base64.b64encode(dill.dumps(value)).decode()

            # 其余直接赋值
            else:
                request_dict[key] = value

        return request_dict

    def to_json(self, **kwargs) -> json:
        """
        转化为 json 字符串

        :param kwargs: json 参数
        :return:
        """

        kwargs.setdefault('ensure_ascii', False)

        return json.dumps(self.to_dict(), **kwargs)

    @property
    def domain(self) -> str:
        """
        获取请求的域名

        :return:
        """
        return urlparse(self.url).netloc

    def __getattr__(self, item):
        """
        可以 request.xxx 进行访问，但是这里严格一点没有就报错避免误导

        :param item:
        :return:
        """
        if item in self.__dict__:
            return self.__dict__[item]

        raise AttributeError(f'未定义的属性：{item}')

    def __setattr__(self, key, value):
        """
        实现 requests.xxx 设置参数（只有 self.xxx 才会进入该函数）

        @param key:
        @param value:
        @return:
        """
        self.__dict__[key] = value

        if key in self.__class__.__REQUEST_ATTRS__:
            self._requests_params[key] = value

    def __lt__(self, other):
        """
        返回比较，否则使用优先级队列会报错

        :param other:
        :return:
        """
        return self.priority < other.priority

    def __str__(self):
        return f"<Request {self.method}-{self.url}>"


class LoadRequest:
    """
        从字典加载出 请求
    """

    @classmethod
    def load_from_dict(cls, **kwargs) -> Request:
        """
        从字典导入

        :param kwargs: request 参数
        """
        for key, value in kwargs.items():
            if key in pickle_attr:
                kwargs[key] = dill.loads(base64.b64decode(value))

        return Request(**kwargs)

    @classmethod
    def load_from_json(cls, data: str) -> Request:
        """
        从 json 字符串导入

        :return:
        """

        return cls.load_from_dict(**json.loads(data))



"""
    报警模块

    使用方式见 gsender 模块对应源码
"""
from palp import settings
from typing import Union, List
from gsender import EmailSender, DingTalkSender


def send_email(
        receiver: Union[List[str], str],
        subject: str = 'Palp 报警系统',
        content_text: str = None,
        content_html: Union[List[str], str] = None,
        files: Union[List[str], str] = None,
        headers: dict = None
):
    """
    发送邮件

    :param receiver: 接收者
    :param subject: 主题
    :param content_text: 文本内容（二选一）
    :param content_html: html 内容（二选一）
    :param files: 文件路径或路径列表
    :param headers: 额外标题
    :return:
    """
    email_sender = EmailSender(user=settings.EMAIL_USER, pwd=settings.EMAIL_PWD)

    return email_sender.send(
        receiver=receiver,
        subject=subject,
        content_text=content_text,
        content_html=content_html,
        files=files,
        headers=headers
    )


def send_dingtalk() -> DingTalkSender:
    """
    钉钉群聊发送器，支持多种消息类型

    :return:
    """
    return DingTalkSender(secret=settings.DING_TALK_SECRET, access_token=settings.DING_TALK_ACCESS_TOKEN)



"""
    spider 基类

    注意：要明白类的赋值操作、修改操作的概念
    赋值操作：给类本身进行赋值
    修改操作：类本身赋值后，是基于类本身修改，类本身未赋值，那么就会向父类进行查找，修改的就是父类
"""
import sys
import time
import uuid
import types
import inspect
import threading
from pathlib import Path
from palp import settings
from loguru import logger
from abc import abstractmethod
from typing import Union, List
from gsender import EmailSender
from palp.network.request import Request
from palp.network.response import Response
from requests.cookies import RequestsCookieJar
from palp.controller import SpiderController, ItemController
from palp.tool.short_module import sort_module, import_module
from palp.decorator.decorator_run_func_by_thread import RunByThreadDecorator


class SpiderBase(threading.Thread):
    spider_name = None  # 用户输入的名字
    spider_domains = []  # 允许 spider 通过的域名，默认都可以
    spider_settings: Union[types.ModuleType, dict] = None  # 用来指定爬虫单独的配置，可以是一个引入的 settings.py 模块 或 字典
    SPIDER_MIDDLEWARE: List[types.ModuleType] = []

    def __new__(cls, *args, **kwargs):
        """
        实例化之前加载设置

        :param args:
        :param kwargs:
        """
        if not cls.SPIDER_MIDDLEWARE:
            cls.from_settings()

        return object.__new__(cls)

    @classmethod
    def from_settings(cls) -> None:
        """
        spider 启动时的自定义配置解析自定义的配置文件

        :return:
        """
        # 修改日志
        if settings.LOG_SAVE:
            log_path = Path(settings.LOG_PATH).joinpath(cls.__name__)
            log_path.mkdir(parents=True, exist_ok=True)

            logger.remove()
            logger.add(
                sink=log_path.joinpath(f'{cls.__name__}.log'),
                enqueue=True,
                backtrace=True,
                diagnose=True,
                rotation=settings.LOG_ROTATION,
                compression=settings.LOG_COMPRESSION,
                level=settings.LOG_LEVEL,
                retention=settings.LOG_RETENTION,
            )

            # 让日志同时能够输出
            if settings.LOG_SHOW:
                logger.add(sys.stdout, level=settings.LOG_LEVEL)

        # 处理每个爬虫单独的设置
        if isinstance(cls.spider_settings, types.ModuleType):
            for key, value in cls.spider_settings.__dict__.items():
                if not key.startswith('__'):
                    setattr(settings, key, value)
        elif isinstance(cls.spider_settings, dict):
            for key, value in cls.spider_settings.items():
                setattr(settings, key, value)
        elif cls.spider_settings is None:
            pass
        else:
            logger.warning("不受支持的设置格式，仅支持模块或者字典！")

        # 引入中间件
        modules = sort_module(
            cls_dict=settings.SPIDER_MIDDLEWARE,
            palp_cls_dict=settings.PALP_SPIDER_MIDDLEWARE
        )

        cls.SPIDER_MIDDLEWARE = import_module(modules)

        # 创建连接，检查连接，防止连接不可用
        from palp import conn

        if settings.SPIDER_TYPE != 1 and conn.redis_conn is None:
            raise ConnectionError('redis 未连接！')
        if conn.redis_conn:
            conn.redis_conn.info()
        if settings.MYSQL_HOST:
            conn.mysql_conn.connection()
        if settings.PG_HOST:
            conn.pg_conn.connection()
        if settings.MONGO_HOST:
            conn.mongo_conn.conn.server_info()

        # 检查其余可用性
        # 邮件是否可用
        if settings.EMAIL_USER and settings.EMAIL_PWD:
            EmailSender(user=settings.EMAIL_USER, pwd=settings.EMAIL_PWD).check_availability()

    def __init__(self, thread_count: int = None, request_filter: bool = False, item_filter: bool = False):
        super().__init__()
        # 加载快捷的设置
        if thread_count:
            setattr(settings, 'REQUEST_THREADS', thread_count)
        if request_filter:
            setattr(settings, 'REQUEST_FILTER', request_filter)
        if item_filter:
            setattr(settings, 'ITEM_FILTER', item_filter)

        self.queue = None  # 请求队列
        self.queue_item = None  # item 队列
        self.spider_done = False  # 爬虫是否运行结束
        self.item_controller_list = []  # 存储所有的解析器
        self.spider_controller_list = []  # 存储所有的解析器
        self.distribute_thread_list = []  # 存储所有任务分发线程

        self.spider_record = {'all': 0, 'failed': 0, 'succeed': 0}  # 爬取情况记录
        self.spider_uuid = uuid.uuid1().hex  # uuid 用作线上区分

    def start_requests(self) -> None:
        """
        初始请求的发出

        @result: yield Request()
        """

    def parse(self, request: Request, response: Response) -> None:
        """
        默认的解析函数（仅 start_requests 默认的）

        @result:
        """

    def wait_distribute_thread_done(self) -> None:
        """
        等待所有任务分发线程执行结束

        :return:
        """
        while True:
            for distribute_thread in self.distribute_thread_list.copy():
                if not distribute_thread.is_alive():
                    self.distribute_thread_list.remove(distribute_thread)

            if len(self.distribute_thread_list) == 0:
                break

            time.sleep(0.1)

    def wait_spider_controller_done(self) -> None:
        """
        等待线程执行结束

        :return:
        """
        while True:
            if self.all_spider_controller_is_waiting():
                logger.debug("所有线程都已挂起，即将停止")
                self.stop_all_spider_controller()
                break
            elif self.all_spider_controller_is_done():
                break

            time.sleep(0.1)  # 不加延迟将会导致性能问题

    def wait_item_controller_done(self) -> None:
        """
        等待 item buffer 结束

        :return:
        """
        while True:
            for item_controller in self.item_controller_list.copy():
                if not item_controller.is_alive():
                    self.item_controller_list.remove(item_controller)

            if len(self.item_controller_list) == 0:
                break

            time.sleep(0.1)

    def stop_all_spider_controller(self) -> None:
        """
        停止所有 spider_controller 运行

        :return:
        """
        for p in self.spider_controller_list:
            p.stop = True

        while True:
            if self.all_spider_controller_is_done():
                break

            time.sleep(0.1)

    def all_distribute_thread_is_done(self) -> bool:
        """
        是否分发任务已经完成

        :return:
        """
        done_status = True

        for p in self.distribute_thread_list:
            if p.is_alive():
                done_status = False
                break

        return done_status

    def all_spider_controller_is_waiting(self) -> bool:
        """
        是否所有 spider_controller 都没事干

        :return:
        """
        waiting_status = True

        for p in self.spider_controller_list.copy():
            if not p.is_alive():
                self.spider_controller_list.remove(p)  # 避免死线程导致程序无法退出
                continue

            if not p.waiting:
                waiting_status = False
                break

        return waiting_status

    def all_spider_controller_is_done(self) -> bool:
        """
        判断所有 spider_controller 是否已经结束

        :return:
        """

        done_status = True

        for p in self.spider_controller_list:
            if p.is_alive():
                done_status = False
                break

        if done_status:
            self.spider_done = True

        return done_status

    @RunByThreadDecorator(daemon=True)
    def start_distribute(self) -> None:
        """
        刚开始的分发任务

        :return:
        """
        # 检查是否是 yield 函数，不是不报错，避免只是执行失败的
        if inspect.isgeneratorfunction(self.start_requests):
            # 检查起始函数发出请求
            for request in self.start_requests():
                if not isinstance(request, Request):
                    raise ValueError("start_requests 仅支持 yield Request")

                # 起始函数无 callback 默认添加
                if request.callback is None:
                    request.callback = self.parse

                # 先改成名字，不然后续无法序列化
                request.callback = request.callback.__name__

                # 为每一个起始函数添加一个 cookie_jar
                request.cookie_jar = RequestsCookieJar()

                self.queue.put(request)
        else:
            logger.warning("self.start_requests 函数不是生成器函数，将不会执行！")

    def start_controller(self) -> None:
        """
        启动 controller

        :return:
        """
        # 启动相应数量的爬虫
        for _ in range(settings.REQUEST_THREADS):
            controller = SpiderController(q=self.queue, q_item=self.queue_item, spider=self)
            self.spider_controller_list.append(controller)
            controller.start()

        # 启动相应数量的item
        for _ in range(settings.ITEM_THREADS):
            controller = ItemController(q=self.queue_item, spider=self)
            self.item_controller_list.append(controller)
            controller.start()

    @abstractmethod
    def run(self) -> None:
        """
        程序启动入口

        :return:
        """

    @property
    def name(self) -> str:
        """
        获取 spider 的名字

        :return:
        """

        return self.__class__.spider_name or self.__class__.__name__

    def __setattr__(self, key, value):
        """
        设置属性

        :param key:
        :param value:
        :return:
        """
        self.__dict__[key] = value

    def __getattr__(self, key):
        """
        获取属性

        :param key:
        :return:
        """
        if key not in self.__dict__:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{key}'")

        return self.__dict__[key]


class Spider(SpiderBase):
    """
        外部引用：Spider 基类
    """

    def parse(self, request, response) -> None:
        """
        默认的解析函数（仅 start_requests 默认的）

        @result:
        """
        pass

    @abstractmethod
    def run(self) -> None:
        """
        程序启动入口

        :return:
        """



from palp.spider.spider_jump import JumpSpider
from palp.spider.spider_local import LocalSpider
from palp.spider.spider_cycle import CycleSpider
from palp.spider.spider_distributive import DistributiveSpider



"""
    jump spider（本地执行，用于处理比较频繁的循环事件，完全独立正常的 spider）

    jump spider 内置 spider_start、spider_close
    jump spider 参数 request_middleware 可传请求中间件

    注意：
        jump spider 不走 spider、request 中间件、不走 记录装饰器
        jump spider 不能 yield item yield 的都不会处理
        jump spider 最后会执行 jump_out 合并 cookie、meta，可以自定义
"""
import inspect
from palp import settings
from abc import abstractmethod
from typing import Union, List, Callable
from palp.network.request import Request
from palp.network.response import Response
from requests.cookies import RequestsCookieJar
from palp.tool.short_module import import_module
from palp.exception import NotGeneratorFunctionError
from palp.controller.controller_spider_jump import JumpController


class JumpSpider:
    """
        jump spider（用于处理比较频繁的循环事件）

        注意：不走 spider 中间件，不处理 item
    """

    def __init__(self, request: Request, request_middleware: Union[List[Callable], Callable] = None, **kwargs):
        """

        :param request: jump 之后需要执行的请求
        :param request_middleware: 请求中间件
        :param kwargs: 需要被 self.xxx 访问到的参数
        """
        self.queue = import_module(settings.REQUEST_QUEUE[1][settings.REQUEST_QUEUE_MODE])[0]  # 请求队列
        self.request = request
        self.request_middleware = request_middleware

        # 导入一下自定义设置
        for key, value in kwargs.items():
            setattr(self, key, value)

    def spider_start(self) -> None:
        """
        spider 开始时的操作

        :param spider:
        :return:
        """

    def spider_close(self) -> None:
        """
        spider 开始时的操作

        :param spider:
        :return:
        """

    @abstractmethod
    def start_requests(self) -> None:
        """
        起始函数

        :return:
        """
        pass

    def start_distribute(self) -> None:
        """
        刚开始的分发任务

        :return:
        """
        # 校验是否是生成器函数
        if not inspect.isgeneratorfunction(self.start_requests):
            raise NotGeneratorFunctionError("start_requests 必须 yield Request!")

        # 检查起始函数发出请求
        for request in self.start_requests():
            if not isinstance(request, Request):
                raise ValueError("start_requests 仅支持 yield Request")

            # 起始函数无 callback 默认添加
            if request.callback is None:
                request.callback = self.parse

            # 为每一个起始函数添加一个 cookie_jar，如果有就不加
            if not request.cookie_jar:
                request.cookie_jar = RequestsCookieJar()

            self.queue.put(request)

    def parse(self, request: Request, response: Response) -> None:
        """
        默认的解析函数（仅 start_requests 默认的）

        @result:
        """

    def run(self) -> None:
        """
        执行 jump spider

        :return:
        """
        self.spider_start()

        try:
            # 分发任务
            self.start_distribute()

            # 执行流程
            JumpController(q=self.queue, request_middleware=self.request_middleware, spider=self).run()
        finally:
            self.spider_close()

    def jump_out(self, request: Request):
        """
        jump spider 执行结束之后，将会调用合并 cookie、meta

        :param request:
        :return:
        """
        self.request.cookie_jar.update(request.cookie_jar)
        if self.request.meta:
            self.request.meta.update(request.meta)
        else:
            self.request.meta = request.meta

    def __setattr__(self, key, value):
        """
        设置属性

        :param key:
        :param value:
        :return:
        """
        self.__dict__[key] = value

    def __getattr__(self, key):
        """
        获取属性

        :param key:
        :return:
        """
        if key not in self.__dict__:
            raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{key}'")

        return self.__dict__[key]
